import importlib
from functools import partial
from socket import socket
from typing import Literal, Type

from cow.manager.process import SanicProcess
from cow.manager.utils import createManager
from cow.web.configs_handler import ConfigsHandler

from luna_lambda_tools import LambdaUserHandler

from .base_application import BaseAppCtx
from .base_application_handlers import HandlersLambdaApp
from .base_application_standalone import StandaloneLambdaApp
from .base_application_tasks import LambdaTasksApp
from .config import DEFAULT_PORT
from .handlers import Handler, HealthCheckHandler, SpecDocHandler
from .helpers import getParsedCMDArgs
from .task_unit_config import ServiceSettings as LambdaTaskServiceSettings
from .unit_config import ServiceSettings as LambdaServiceSettings


def getAppCls(lambdaType: Literal["standalone", "handlers", "tasks"]):
    """Get app class by lambda type"""
    if lambdaType == "tasks":
        appCls = LambdaTasksApp
    elif lambdaType == "handlers":
        appCls = HandlersLambdaApp
    else:
        appCls = StandaloneLambdaApp
    return appCls


def extractAdditionalRoutes() -> list[tuple[str, Type]]:
    """Extract additional routes from lambda"""
    additionalRoutes = getattr(importlib.import_module("lambda_main", "main"), "ROUTES_TO_ADD", [])
    routes = []
    for handler in additionalRoutes:
        if issubclass(handler, LambdaUserHandler) and handler.__module__.endswith("lambda_main"):
            routes.append((handler.route, handler))
    return routes


def addRoutes(application: StandaloneLambdaApp | HandlersLambdaApp | LambdaTasksApp):
    """Add route to app by lambda type"""
    routes = [
        (f"/healthcheck", HealthCheckHandler),
        (f"/docs/spec", SpecDocHandler),
        (f"/config", ConfigsHandler),
    ]
    if isinstance(application, (StandaloneLambdaApp, HandlersLambdaApp)):
        routes.append(("/main", Handler))
    routes.extend(extractAdditionalRoutes())
    application.addRoutes(routes=routes)
    return application


def getSettingsClass(lambdaType: Literal["standalone", "handlers", "tasks"]):
    """Get settings class by lambda type"""
    if lambdaType == "tasks":
        return LambdaTaskServiceSettings
    return LambdaServiceSettings


def _runLambda(sock: socket, lambdaType: Literal["standalone", "handlers", "tasks"]):
    """Run lambda"""
    lambdaName, ctxClass = getLambdaNameNCtx(lambdaType=lambdaType)

    application = getAppCls(lambdaType=lambdaType)
    application.ctxCls = ctxClass
    application = application()
    addRoutes(application=application)
    application.run(sock=sock)


def getCtxClass(
    lambdaType: Literal["standalone", "handlers", "tasks"], lambdaVersion: int, lambdaName: str
) -> object | None:
    """Get ctx class by lambda type"""
    if lambdaType == "standalone":
        from .base_application_standalone import AppCtx
    elif lambdaType == "handlers":
        from .base_application_handlers import AppCtx
    else:
        from .base_application_tasks import AppCtx

    AppCtx.lambdaVersion = lambdaVersion
    AppCtx.lambdaName = lambdaName
    return AppCtx


def getLambdaNameNCtx(lambdaType: Literal["standalone", "handlers", "tasks"]) -> tuple[str, Type[BaseAppCtx]]:
    """Get lambda name and context"""
    cmdArgs = getParsedCMDArgs(lambdaType=lambdaType)
    lambdaName = cmdArgs.name
    if cmdArgs.name.startswith("lambda-"):
        lambdaName = cmdArgs.name[len("lambda-") :]
    ctxClass = getCtxClass(lambdaType=lambdaType, lambdaVersion=cmdArgs.version, lambdaName=lambdaName)
    return lambdaName, ctxClass


def run(lambdaType: Literal["standalone", "handlers", "tasks"]):
    """Run lambda using process manager"""
    lambdaName, ctxClass = getLambdaNameNCtx(lambdaType=lambdaType)

    target = partial(_runLambda, lambdaType=lambdaType)

    processes = [SanicProcess(name=lambdaName, target=target)]
    manager = createManager(
        processes,
        configClass=getSettingsClass(lambdaType=lambdaType),
        ctxClass=ctxClass,
        port=DEFAULT_PORT,
    )

    manager.run()
